@@ -34,20 +34,21 @@ class VisionClassifierTrainer:
   def __init__(
     self,
     ids2labels,
+    output_dir        :str,
+    model_name        :str,
     model             :torch.nn.Module,
     feature_extractor :FeatureExtractionMixin,
     dataset           :torch.utils.data.Dataset,
-    model_name        :str,
-    output_dir        :str,
-    lr           = 2e-5,
-    batch_size   = 8,
-    max_epochs   = 1,
-    cores        = 4,
-    eval_metric  = "accuracy",
-    shuffle      = True,
-    fp16         = True,
-    balanced     = False,
-    augmentation = False,
+    max_epochs    = 1,
+    cores         = 4,
+    batch_size    = 8,
+    test_ratio    = 0.15,
+    lr            = 2e-5,
+    eval_metric   = "accuracy",
+    fp16          = True,
+    shuffle       = True,
+    balanced      = False,
+    augmentation  = False,
   ):
 
     self.model_name        = model_name
@@ -57,6 +58,7 @@ class VisionClassifierTrainer:
     self.batch_size        = batch_size
     self.max_epochs        = max_epochs
     self.shuffle           = shuffle
+    self.test_ratio        = test_ratio
     self.model             = model
     self.feature_extractor = feature_extractor
     self.cores             = cores
@@ -77,36 +79,39 @@ class VisionClassifierTrainer:
 
     # Get the model output path
     self.output_path = self.__getOutputPath()
-    self.logs_path = self.output_path
+    self.logs_path   = self.output_path
     
     # Open the logs file
     self.__openLogs()
 
+    # Split and convert to dataloaders
+    self.train, self.test = self.__splitDatasets()
+
     """
     🏗️ Build the trainer
     """
     self.training_args = TrainingArguments(
         output_dir                  = self.output_path,
-        save_steps                  = 10000,
         save_total_limit            = 2,
         weight_decay                = 0.01,
+        save_steps                  = 10000,
         learning_rate               = self.lr,
         per_device_train_batch_size = self.batch_size,
         per_device_eval_batch_size  = self.batch_size,
         num_train_epochs            = self.max_epochs,
         metric_for_best_model       = self.eval_metric,
         logging_dir                 = self.logs_path,
-        fp16                        = self.fp16,
         evaluation_strategy         = "epoch",
-        overwrite_output_dir        = True,
         load_best_model_at_end      = False,
+        overwrite_output_dir        = True,
+        fp16=self.fp16,
     )
     
     self.trainer = Trainer(
       self.model,
       self.training_args,
-      train_dataset = self.dataset["train"],
-      eval_dataset  = self.dataset["test"],
+      train_dataset = self.train,
+      eval_dataset  = self.test,
       data_collator = self.collator,
     )
     print("Trainer builded!")
@@ -149,28 +154,125 @@ class VisionClassifierTrainer:
     return path
 
   """
-  🧪 Evaluate the performances of the system of the test sub-dataset
+  🧬 Apply data augmentation on the input image
+  Source: https://medium.com/lunit/photometric-data-augmentation-in-projection-radiography-bed3ae9f55c3
   """
-  def __evaluate(self):
-        
-    all_preds  = []
-    all_target = []
+  def __augmentation(self, image, beta=0.33):
+
+    # Random augmentation
+    if random.randint(0,100) < (beta*100):
+    
+      # Random Contrast
+      im3 = ImageEnhance.Contrast(image)
+      im3.enhance(random.uniform(0.5, 1.0)).show()
+      
+      # Random Noise
+
+    return image
+
+  """
+  ⚖️ Balance the dataset according to the less represented label
+  """
+  def __balance(self, train_ds, train_classes):
+
+    # Get the less represented label in train
+    less_represented_train = min(train_classes)
+
+    labels = {}
 
     # For each image
-    for image, label in tqdm(self.dataset["test"]):
+    for img, label in train_ds:
 
-        # Compute
-        inputs  = self.feature_extractor(images=image, return_tensors="pt").to(self.device)
-        outputs = self.model(**inputs)
+      # Create the label array if isn't
+      if label not in labels:
+        labels[label] = []
+      
+      # Add the image to the array
+      labels[label].append(img)
+    
+    # New dataset
+    balanced_ds = []
 
-        # Get predictions from the softmax layer
-        preds = outputs.logits.softmax(1).argmax(1).tolist()
-        all_preds.extend(preds)
+    # For each label
+    for label in labels:
 
-        # Get hypothesis
-        all_target.append(label)
+      # Get images
+      imgs = labels[label]
 
-    return all_preds, all_target
+      # For each image
+      for img in imgs[0:less_represented_train]:
+
+        # Create a tuple: image and label 
+        t = (img, label)
+
+        # Add it
+        balanced_ds.append(t)
+
+    print("The less represented label in train as " + str(less_represented_train) + " occurrences")
+    print("Size of train after balancing is " + str(len(balanced_ds)))
+
+    return balanced_ds
+    
+  """
+  ✂️ Split the dataset into sub-datasets
+  """
+  def __splitDatasets(self):
+
+    print("Split Datasets...")
+    
+    indices = torch.randperm(len(self.dataset)).tolist()
+
+    # Index of the validation corpora
+    train_index = math.floor(len(indices) * (1 - self.test_ratio))
+
+    # TRAIN
+    train_ds = torch.utils.data.Subset(self.dataset, indices[:train_index])
+    ct_train = Counter([label for _, label in train_ds])
+    train_classes = [ct_train[a] for a in sorted(ct_train)]
+    
+    # If balanced is enabled
+    if self.balanced == True:
+
+      # Balance the train dataset
+      train_ds = self.__balance(train_ds, train_classes)
+      
+      # Compute again the stats
+      ct_train = Counter([label for _, label in train_ds])
+      train_classes = [ct_train[a] for a in sorted(ct_train)]
+
+    # If data augmentation is enabled
+    if self.augmentation == True:
+
+      new_ds = []
+
+      # For each annotated image
+      for img, label in train_ds:
+
+        # Augment it
+        new_ds.append(
+          (self.__augmentation(img), label)
+        )
+      
+      # Replace by the augmented data
+      train_ds = new_ds
+
+    # TEST
+    test_ds = torch.utils.data.Subset(self.dataset, indices[train_index:])
+    ct_test = Counter([label for _, label in test_ds])
+    test_classes = [ct_test[a] for a in sorted(ct_test)]
+
+    table_repartition = [
+      ["Train"] + train_classes + [str(len(train_ds))],
+      ["Test"] + test_classes + [str(len(test_ds))],
+    ]
+
+    repartitions_table = tabulate(table_repartition, ["Dataset"] + list(self.ids2labels.values()) + ["Total"], tablefmt="pretty")
+    print(repartitions_table)
+
+    # Write logs
+    self.logs_file.write(repartitions_table + "\n")
+
+    return train_ds, test_ds
 
   """
   🧪 Evaluate the performances of the system of the test sub-dataset given a f1-score
@@ -191,24 +293,24 @@ class VisionClassifierTrainer:
     # Add macro scores for each classes
     for i in range(len(labels)):
         table_f_score.append(
-            [
-                self.ids2labels[labels[i]],
-                str(round(precision[i] * 100,2)) + " %",
-                str(round(recall[i] * 100,2)) + " %",
-                str(round(fscore[i] * 100,2)) + " %",
-                support[i],
-            ]
+          [
+              self.ids2labels[labels[i]],
+              str(round(precision[i] * 100,2)) + " %",
+              str(round(recall[i] * 100,2)) + " %",
+              str(round(fscore[i] * 100,2)) + " %",
+              support[i],
+          ]
         )
 
     # Add global macro scores
     table_f_score.append(
-        [
-            "Macro",
-            str(round((sum(precision) / len(precision)) * 100, 2)) + " %",
-            str(round((sum(recall)    / len(recall))    * 100, 2)) + " %",
-            str(round((sum(fscore)    / len(fscore))    * 100, 2)) + " %",
-            str(sum(support)),
-        ]
+      [
+          "Macro",
+          str(round((sum(precision) / len(precision)) * 100, 2)) + " %",
+          str(round((sum(recall)    / len(recall))    * 100, 2)) + " %",
+          str(round((sum(fscore)    / len(fscore))    * 100, 2)) + " %",
+          str(sum(support)),
+      ]
     )
 
     # Print precision, recall, f-score and support for each classes
@@ -224,6 +326,36 @@ class VisionClassifierTrainer:
 
     return all_target, all_preds
 
+  """
+  🧪 Evaluate the performances of the system of the test sub-dataset for the chosen evaluation metric
+  """
+  def evaluate(self):
+    return self.trainer.evaluate()
+
+  """
+  🧪 Evaluate the performances of the system of the test sub-dataset
+  """
+  def __evaluate(self):
+        
+    all_preds  = []
+    all_target = []
+
+    # For each image
+    for image, label in tqdm(self.test):
+
+        # Compute
+        inputs  = self.feature_extractor(images=image, return_tensors="pt").to(self.device)
+        outputs = self.model(**inputs)
+
+        # Get predictions from the softmax layer
+        preds = outputs.logits.softmax(1).argmax(1).tolist()
+        all_preds.extend(preds)
+
+        # Get hypothesis
+        all_target.append(label)
+
+    return all_preds, all_target
+
   """
   🧪 Test on a single image
   """
